/*!
 * This file defines the class that is used to represent SC collections
 * in Objective-C.
 */

#import <Foundation/Foundation.h>
#import "PyrObject.h"

/*
 * Instances of this class are used as proxies for SC Lists and
 * ArrayedCollections when these are passed to Objective-C code. Because 
 * this class is a subclass of NSMutableArray these SC collections 
 * can be used everywhere NSArray or NSMutableArray objects 
 * are expected. This is done in a single class (instead of an NSArray
 * subclass for ArrayedCollections and an NSMutableArray subclass for
 * lists) because ArrayedCollections in SC have a .add method and grow 
 * towards the size they were created with. In other words, NSMutableArray
 * methods need to work on both lists and ArrayedCollections. An exception
 * is raised if addObject: or insertObject:atIndex: is called on an
 * ArrayedColection that is at its maximum size.
 */
@interface CCArray : NSMutableArray
{
	PyrObject *scArray, *scList;
}

/*!
 * @method newWithSCArray:
 * @abstract Create a new CCArray for a specific SC ArrayedCollection
 * @param value A SC ArrayedCollection
 * @result Returns an autoreleased instance representing value
 */
+ (CCArray *)newWithSCArray:(PyrObject *)aValue;

/*!
 * @method initWithSCArray:
 * @abstract Initialise a CCArray for a specific SC ArrayedCollection
 * @param value A SC ArrayedCollection
 * @result Returns self
 */
- (CCArray *)initWithSCArray:(PyrObject *)aValue;

/*!
 * @method newWithSCList:
 * @abstract Create a new CCArray for a specific SC List
 * @param value A SC List
 * @result Returns an autoreleased instance representing value
 */
+ (CCArray *)newWithSCList:(PyrObject *)aValue;

/*!
 * @method initWithSCList:
 * @abstract Initialise a CCArray for a specific SC List
 * @param value A SC List
 * @result Returns self
 */
- (CCArray *)initWithSCList:(PyrObject *)aValue;

/*!
 * @method dealloc
 * @abstract Access the wrapped SC collection
 * @result  Returns a new reference to the wrapped SC collection.
 */
-(PyrObject *)pyrObject;

/*!
 * @method count
 * @result  Returns the length of the wrapped SC collection
 */
-(int)count;

/*!
 * @method objectAtIndex:
 * @param idx An index
 * @result  Returns the object at the specified index in the wrapped SC
 *          collection
 */
- (id)objectAtIndex:(int)idx;

/*!
 * @method replaceObjectAtIndex:withObject:
 * @abstract Replace the current value at idx by the new value
 * @discussion This method will raise an exception when the wrapped SC
 *             collection is immutable.
 * @param idx An index
 * @param newValue A replacement value
 */
-(void)replaceObjectAtIndex:(int)idx withObject:(id)newValue;

/*!
 * @method getObjects:inRange:
 * @abstract Fetch objects in the specified range
 * @discussion The output buffer must have enough space to contain all
 *             requested objects, the range must be valid.
 *
 *             This method is not documented in the NSArray interface, but
 *             is used by Cocoa on MacOS X 10.3 when an instance of this
 *             class is used as the value for -setObject:forKey: in
 *             NSUserDefaults.
 * @param buffer  The output buffer
 * @param range   The range of objects to fetch.
 */
-(void)getObjects:(id*)buffer inRange:(NSRange)range;

/*!
 * @method addObject:
 * @abstract Append an object
 * @param anObject The object that will be append
 */
-(void)addObject:(id)anObject;

/*!
 * @method insertObject:atIndex:
 * @abstract Insert an object at the specified index
 * @param anObject The object to insert
 * @param idx  The index
 */
-(void)insertObject:(id)anObject atIndex:(unsigned)idx;

/*!
 * @method removeLastObject
 * @abstract Remove the last object of the collection
 */
-(void)removeLastObject;

/*!
 * @method removeObjectAtIndex:
 * @abstract Remove a specific item
 * @param idx Which object should be removed
 */
-(void)removeObjectAtIndex:(unsigned)idx;

@end
